"""Commands for displaying dataset overlay information."""

import json
import sys

import click

import pygments
import pygments.lexers
import pygments.formatters

import dtoolcore

from dtool_cli.cli import (
    dataset_uri_argument,
)


@click.group()
def overlay():
    """
    Get information about item metadata stored in overlays.
    """


@overlay.command()
@dataset_uri_argument
def ls(dataset_uri):
    """
    List the overlays in the dataset.
    """
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    for overlay_name in dataset.list_overlay_names():
        click.secho(overlay_name)


@overlay.command()
@dataset_uri_argument
@click.argument("overlay_name")
def show(dataset_uri, overlay_name):
    """
    Show the content of a specific overlay.
    """
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    try:
        overlay = dataset.get_overlay(overlay_name)
    except:  # NOQA
        click.secho(
            "No such overlay: {}".format(overlay_name),
            fg="red",
            err=True
        )
        sys.exit(11)

    formatted_json = json.dumps(overlay, indent=2)
    colorful_json = pygments.highlight(
        formatted_json,
        pygments.lexers.JsonLexer(),
        pygments.formatters.TerminalFormatter())
    click.secho(colorful_json, nl=False)



"""Commands for getting information about datasets."""

import sys

import click

import pygments
import pygments.lexers
import pygments.formatters

import dtoolcore
from dtoolcore.compare import (
    diff_identifiers,
    diff_sizes,
    diff_content,
)

from dtool_cli.cli import (
    dataset_uri_argument,
    dataset_uri_validation,
    CONFIG_PATH,
)

from dtool_info.utils import sizeof_fmt, date_fmt

item_identifier_argument = click.argument("item_identifier")


@click.command()
@click.option(
    "-f",
    "--full",
    is_flag=True,
    help="Include file hash comparisons."
)
@dataset_uri_argument
@click.argument("reference_dataset_uri", callback=dataset_uri_validation)
def diff(full, dataset_uri, reference_dataset_uri):
    """Report the difference between two datasets.

    1. Checks that the identifiers are identicial
    2. Checks that the sizes are identical
    3. Checks that the hashes are identical, if the '--full' option is used

    If a differences is detected in step 1, steps 2 and 3 will not be carried
    out. Similarly if a difference is detected in step 2, step 3 will not be
    carried out.

    When checking that the hashes are identical the hashes for the first
    dataset are recalculated using the hashing algorithm of the reference
    dataset.
    """

    def echo_header(desc, ds_name, ref_ds_name, prop):
        click.secho("Different {}".format(desc), fg="red")
        click.secho("ID, {} in '{}', {} in '{}'".format(
            prop, ds_name, prop, ref_ds_name))

    def echo_diff(diff):
        for d in diff:
            line = "{}, {}, {}".format(d[0], d[1], d[2])
            click.secho(line)

    ds = dtoolcore.DataSet.from_uri(dataset_uri)
    ref_ds = dtoolcore.DataSet.from_uri(reference_dataset_uri)

    num_items = len(list(ref_ds.identifiers))

    ids_diff = diff_identifiers(ds, ref_ds)
    if len(ids_diff) > 0:
        echo_header("identifiers", ds.name, ref_ds.name, "present")
        echo_diff(ids_diff)
        sys.exit(1)

    with click.progressbar(length=num_items,
                           label="Comparing sizes") as progressbar:
        sizes_diff = diff_sizes(ds, ref_ds, progressbar)
    if len(sizes_diff) > 0:
        echo_header("sizes", ds.name, ref_ds.name, "size")
        echo_diff(sizes_diff)
        sys.exit(2)

    if full:
        with click.progressbar(length=num_items,
                               label="Comparing hashes") as progressbar:
            content_diff = diff_content(ds, ref_ds, progressbar)
        if len(content_diff) > 0:
            echo_header("content", ds.name, ref_ds.name, "hash")
            echo_diff(content_diff)
            sys.exit(3)


def _list_dataset_items(uri, quiet, verbose):
    try:
        dataset = dtoolcore.DataSet.from_uri(
            uri=uri,
            config_path=CONFIG_PATH
        )
    except dtoolcore.DtoolCoreTypeError:
        click.secho(
            "Cannot list the items of a proto dataset",
            fg="red",
            err=True
        )
        sys.exit(1)
    for i in dataset.identifiers:
        props = dataset.item_properties(i)
        line = "{}  {}".format(i, props["relpath"])
        if verbose:
            line = "{}{}  {}".format(
                i,
                sizeof_fmt(props["size_in_bytes"]),
                props["relpath"]
            )
        if quiet:
            line = props["relpath"]
        click.secho(line)


def _list_datasets(base_uri, quiet, verbose):
    base_uri = dtoolcore.utils.sanitise_uri(base_uri)
    StorageBroker = dtoolcore._get_storage_broker(base_uri, CONFIG_PATH)
    info = []
    for uri in StorageBroker.list_dataset_uris(base_uri, CONFIG_PATH):
        admin_metadata = dtoolcore._admin_metadata_from_uri(uri, CONFIG_PATH)
        fg = "green"
        name = admin_metadata["name"]
        if admin_metadata["type"] == "protodataset":
            fg = "red"
            name = "*" + name
        i = dict(
            name=name,
            uuid=admin_metadata["uuid"],
            creator=admin_metadata["creator_username"],
            uri=uri,
            fg=fg)
        if "frozen_at" in admin_metadata:
            i["date"] = date_fmt(admin_metadata["frozen_at"])
        info.append(i)

    if len(info) == 0:
        sys.exit(0)

    for i in info:
        click.secho(i["name"], fg=i["fg"])
        if quiet:
            continue
        click.secho("  " + i["uri"])
        if verbose:
            click.secho("  " + i["creator"], nl=False)
            if "date" in i:
                click.secho("  " + i["date"], nl=False)
            click.secho("  " + i["uuid"])


@click.command()
@click.option("-q", "--quiet", is_flag=True)
@click.option("-v", "--verbose", is_flag=True)
@click.argument("uri")
def ls(quiet, verbose, uri):
    """List datasets / items in a dataset.

    If the URI is a dataset the items in the dataset will be listed.
    It is not possible to list the items in a proto dataset.

    If the URI is a location containing datasets the datasets will be listed.
    Proto datasets are highlighted in red.
    """
    if dtoolcore._is_dataset(uri, CONFIG_PATH):
        _list_dataset_items(uri, quiet, verbose)
    else:
        _list_datasets(uri, quiet, verbose)


@click.command()
@dataset_uri_argument
def identifiers(dataset_uri):
    """List the item identifiers in the dataset."""
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    for i in dataset.identifiers:
        click.secho(i)


@click.command()
@dataset_uri_argument
def summary(dataset_uri):
    """Report summary information about a dataset."""
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    creator_username = dataset._admin_metadata["creator_username"]
    frozen_at = dataset._admin_metadata["frozen_at"]
    num_items = len(dataset.identifiers)
    tot_size = sum([dataset.item_properties(i)["size_in_bytes"]
                    for i in dataset.identifiers])

    json_lines = [
        '{',
        '  "name": "{}",'.format(dataset.name),
        '  "uuid": "{}",'.format(dataset.uuid),
        '  "creator_username": "{}",'.format(creator_username),
        '  "number_of_items": {},'.format(num_items),
        '  "size_in_bytes": {},'.format(tot_size),
        '  "frozen_at": {}'.format(frozen_at),
        '}',
    ]
    formatted_json = "\n".join(json_lines)
    colorful_json = pygments.highlight(
        formatted_json,
        pygments.lexers.JsonLexer(),
        pygments.formatters.TerminalFormatter())
    click.secho(colorful_json, nl=False)


@click.group()
def item():
    """
    Get information about an item in the dataset.
    """


@item.command()
@dataset_uri_argument
@item_identifier_argument
def properties(dataset_uri, item_identifier):
    """Report item properties."""
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    try:
        props = dataset.item_properties(item_identifier)
    except KeyError:
        click.secho(
            "No such item in dataset: {}".format(item_identifier),
            fg="red",
            err=True
        )
        sys.exit(20)

    json_lines = [
        '{',
        '  "relpath": "{}",'.format(props["relpath"]),
        '  "size_in_bytes": {},'.format(props["size_in_bytes"]),
        '  "utc_timestamp": {},'.format(props["utc_timestamp"]),
        '  "hash": "{}"'.format(props["hash"]),
        '}',
    ]
    formatted_json = "\n".join(json_lines)
    colorful_json = pygments.highlight(
        formatted_json,
        pygments.lexers.JsonLexer(),
        pygments.formatters.TerminalFormatter())
    click.secho(colorful_json, nl=False)


@item.command()
@dataset_uri_argument
@item_identifier_argument
def fetch(dataset_uri, item_identifier):
    """Return abspath to file with item content.

    Fetches the file from remote storage if required.
    """
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    click.secho(dataset.item_content_abspath(item_identifier))


@click.command()
@click.option(
    "-f",
    "--full",
    is_flag=True,
    help="Include file hash comparisons."
)
@dataset_uri_argument
def verify(full, dataset_uri):
    """Return abspath to file with item content.

    Fetches the file from remote storage if required.
    """
    dataset = dtoolcore.DataSet.from_uri(dataset_uri)
    all_okay = True

    generated_manifest = dataset.generate_manifest()
    generated_identifiers = set(generated_manifest["items"].keys())
    manifest_identifiers = set(dataset.identifiers)

    for i in generated_identifiers.difference(manifest_identifiers):
        message = "Unknown item: {} {}".format(
            i,
            generated_manifest["items"][i]["relpath"]
        )
        click.secho(message, fg="red")
        all_okay = False

    for i in manifest_identifiers.difference(generated_identifiers):
        message = "Missing item: {} {}".format(
            i,
            dataset.item_properties(i)["relpath"]
        )
        click.secho(message, fg="red")
        all_okay = False

    for i in manifest_identifiers.intersection(generated_identifiers):
        generated_hash = generated_manifest["items"][i]["size_in_bytes"]
        manifest_hash = dataset.item_properties(i)["size_in_bytes"]
        if generated_hash != manifest_hash:
            message = "Altered item size: {} {}".format(
                i,
                dataset.item_properties(i)["relpath"]
            )
            click.secho(message, fg="red")
            all_okay = False

    if full:
        for i in manifest_identifiers.intersection(generated_identifiers):
            generated_hash = generated_manifest["items"][i]["hash"]
            manifest_hash = dataset.item_properties(i)["hash"]
            if generated_hash != manifest_hash:
                message = "Altered item hash: {} {}".format(
                    i,
                    dataset.item_properties(i)["relpath"]
                )
                click.secho(message, fg="red")
                all_okay = False

    if not all_okay:
        sys.exit(1)
    else:
        click.secho("All good :)", fg="green")



from setuptools import setup

url = "https://github.com/jic-dtool/dtool-info"
version = "0.10.1"
readme = open('README.rst').read()

setup(
    name="dtool-info",
    packages=["dtool_info"],
    package_data={"dtool_info": ["templates/*"]},
    version=version,
    description="Dtool plugin for getting information about datasets",
    long_description=readme,
    include_package_data=True,
    author="Tjelvar Olsson",
    author_email="tjelvar.olsson@jic.ac.uk",
    url=url,
    install_requires=[
        "click",
        "dtoolcore>=3.0.0",
        "dtool_cli>=0.6.0",
        "jinja2",
        "pygments",
    ],
    entry_points={
        "dtool.cli": [
            "diff=dtool_info.dataset:diff",
            "ls=dtool_info.dataset:ls",
            "summary=dtool_info.dataset:summary",
            "item=dtool_info.dataset:item",
            "identifiers=dtool_info.dataset:identifiers",
            "verify=dtool_info.dataset:verify",
            "overlay=dtool_info.overlay:overlay",
            "inventory=dtool_info.inventory:inventory",
        ],
    },
    download_url="{}/tarball/{}".format(url, version),
    license="MIT"
)



"""Logic for generating inventories of datasets."""

from operator import itemgetter

import click

from jinja2 import Environment, PackageLoader

import dtoolcore

from dtool_cli.cli import CONFIG_PATH

from dtool_info.utils import sizeof_fmt, date_fmt

JINJA2_ENV = Environment(loader=PackageLoader('dtool_info', 'templates'))


def _dataset_info(dataset):
    """Return information about dataset as a dict."""
    info = {}

    info["uri"] = dataset.uri
    info["uuid"] = dataset.uuid

    # Computer and human readable size of dataset.
    tot_size = sum([dataset.item_properties(i)["size_in_bytes"]
                    for i in dataset.identifiers])
    info["size_int"] = tot_size
    info["size_str"] = sizeof_fmt(tot_size)

    info["creator"] = dataset._admin_metadata["creator_username"]
    info["name"] = dataset._admin_metadata["name"]

    info["date"] = date_fmt(dataset._admin_metadata["frozen_at"])

    info["num_items"] = len(dataset.identifiers)

    info["readme_content"] = dataset.get_readme_content()

    return info


def _is_frozen_dataset(uri, CONFIG_PATH):
    admin_metadata = dtoolcore._admin_metadata_from_uri(uri, CONFIG_PATH)
    return admin_metadata["type"] == "dataset"


def _base_uri_info(base_uri):
    StorageBroker = dtoolcore._get_storage_broker(base_uri, CONFIG_PATH)

    info = {}
    info["total_size_int"] = 0
    info["total_items"] = 0
    info["datasets"] = []

    for uri in StorageBroker.list_dataset_uris(base_uri, CONFIG_PATH):
        if not _is_frozen_dataset(uri, CONFIG_PATH):
            continue
        dataset = dtoolcore.DataSet.from_uri(uri)
        dataset_info = _dataset_info(dataset)
        info["datasets"].append(dataset_info)
        info["total_size_int"] += dataset_info["size_int"]
        info["total_items"] += dataset_info["num_items"]

    info["total_size_str"] = sizeof_fmt(info["total_size_int"])
    info["num_datasets"] = len(info["datasets"])

    return info


def _line_formats(info):
    creator_max_len = max(
        [len(ds["creator"]) for ds in info["datasets"]]
    )
    num_items_max_len = max(
        [len(str(ds["num_items"])) for ds in info["datasets"]]
    )

    ds_words = [
        "{size_str}",
        "{" + "creator:{}".format(creator_max_len) + "}",
        "{" + "num_items:{}".format(num_items_max_len) + "}",
        "{date}",
        "{name}",
    ]

    ds_line_format = " ".join(ds_words)

    total_items_offset = creator_max_len + num_items_max_len + 1
    summary_words = [
        "{total_size_str}",
        "{" + "total_items:{}".format(total_items_offset) + "}",
    ]
    summary_line_format = " ".join(summary_words)
    return ds_line_format, summary_line_format


def _cmd_line_report(info):
    # Report on each individual dataset.
    ds_line, summary_line = _line_formats(info)
    for ds_info in sorted(
        info["datasets"],
        key=itemgetter("creator", "date", "name")
    ):
        click.secho(ds_line.format(**ds_info))

    # Summary report on all datasets in base_uri.
    click.secho(summary_line.format(**info))


def _html_report(info):

    template = JINJA2_ENV.get_template("dtool_report.html.j2")
    html = template.render(info)
    click.secho(html)


def _csv_tsv_header_items():
    return [
        "name",
        "size_in_bytes",
        "creator",
        "num_items",
        "date",
        "uri"
    ]


def _csv_tsv_column_items(ds_info):
    return [
        ds_info["name"],
        str(ds_info["size_int"]),
        ds_info["creator"],
        str(ds_info["num_items"]),
        ds_info["date"],
        ds_info["uri"]
    ]


def _csv_tsv_report(info, separator):
    click.secho(separator.join(_csv_tsv_header_items()))
    for ds_info in info["datasets"]:
        click.secho(separator.join(_csv_tsv_column_items(ds_info)))


@click.command()
@click.argument("uri")
@click.option(
    "-f",
    "--format",
    type=click.Choice(["csv", "tsv", "html"]),
    help="Select the output format."
)
def inventory(uri, format):
    """Generate an inventory of datasets in a base URI."""
    base_uri = dtoolcore.utils.sanitise_uri(uri)
    info = _base_uri_info(base_uri)

    if format is None:
        _cmd_line_report(info)
    elif format == "csv":
        _csv_tsv_report(info, ",")
    elif format == "tsv":
        _csv_tsv_report(info, "\t")
    elif format == "html":
        _html_report(info)
